/*
Analyseur syntaxique.

$Id: parser.mly,v 1.1.1.1 1998/06/23 22:35:08 monniaux Exp $
*/

%{
open Syntax;;

(* Donne ((l1, c1),(l2, c2)) avec lx et cx le numro de ligne
   (commenant  1) et de caractre dans la ligne (commenant  0)
   des premier et dernier caractres du bloc actuellement analys. *)
let extent () = (Lexer.location_of_pos (symbol_start ())),
                (Lexer.location_of_pos (pred (symbol_end ())) );;

let item_extent item = (Lexer.location_of_pos (rhs_start item)),
                (Lexer.location_of_pos (pred (rhs_end item)) );;

%}

%token AND ARRAY BEGINBLOCK CASE DECLARE DIV DO ELSE END FOR FROM
%token IF IS MOD NOT OF OR PROCEDURE RECORD REF NEW THEN TO TYPE WHILE
%token ASSIGN INFERIOR SUPERIOR EQUAL IDENTICAL
%token INFERIOR_EQUAL SUPERIOR_EQUAL UNEQUAL UNIDENTICAL 
%token ADD SUB MUL DIV DIV2
%token LPAREN RPAREN LBRACKET RBRACKET SEMICOLON COLON DOT COMMA WEDGE
%token Eof Error

%token <Syntax.iconst> Integer
%token <Syntax.fconst> Float
%token <Syntax.identifier> Identifier

%start bidule
%type <Syntax.statement> bidule

%%

/*
<type>s  = liste  1 ou + lments de type <type> spars par le sparateur
           ad hoc (virgule ou point-virgule)
<type>s0 = liste  0 ou + lements
*/

bidule:
   statement Eof { $1 }
/* Exemple de rcupration */
 | DECLARE error Eof
     {
       let ex = extent() in
         Error.syntax_error "parse error in main program" ex;
         (Expression ErrorExpr), ex
     }


/* Pour les expressions, on n'utilise pas les facilicits intgres
   d'ocamlyacc pour traiter les priorits, on dcalque la grammaire. */
expression:
   name ASSIGN expression { Assignment($1, $3) }
 | disjunction { $1 }

disjunction:
   conjunction { $1 }
 | disjunction OR conjunction { Or($1, $3) }

conjunction:
   comparison { $1 }
 | conjunction AND comparison { And($1, $3) }

comparison:
   relation { $1 }
 | relation EQUAL relation { Equal($1, $3) }
 | relation IDENTICAL relation { Identical($1, $3) }
 | relation UNEQUAL relation { Not(Equal($1, $3)) }
 | relation UNIDENTICAL relation { Not(Identical($1, $3)) }

relation:
   sum { $1 }
 | sum INFERIOR sum { Inferior($1, $3) }
 | sum SUPERIOR sum { Superior($1, $3) }
 | sum INFERIOR_EQUAL sum { Not(Superior($1, $3)) }
 | sum SUPERIOR_EQUAL sum { Not(Inferior($1, $3)) }

sum:
   term { $1 }
 | sum ADD term { Add($1, $3) }
 | sum SUB term { Sub($1, $3) }

term:
   factor { $1 }
 | term MUL factor { Mul($1, $3) }
 | term DIV2 factor { FDiv($1, $3) }
 | term DIV factor { Div($1, $3) }
 | term MOD factor { Mod($1, $3) }

factor:
   primary { $1 }
 | ADD factor { $2 }
 | SUB factor { Minus($2) }
 | NOT factor { Not($2) }

primary:
   Integer { IConst($1) }
 | Float   { FConst($1) }
 | LPAREN expression RPAREN { $2 }
 | block   { Block($1) }
 | IF expression THEN statements0 END { IfThen($2, $4) }
 | IF expression THEN statements0 ELSE statements0 END { IfThenElse($2, $4, $6) }
 | name { Name($1) }

name:
   name LBRACKET arguments RBRACKET { Brackets($1, $3) }
 | name WEDGE { Dereference($1) }
 | name LPAREN arguments0 RPAREN { Call($1, $3) }
 | Identifier { Identifier($1) }

arguments0:
   		{ [] }
 | arguments	{ $1 }

arguments:
   expression	{ [$1] }
 | expression COMMA arguments { $1::$3 }

/* Ceci est un exemple de la faon de rcuprer l'endroit d'o vient
   un lments syntaxique. */
statement_d:
   expression { Expression($1) }
 | FOR Identifier FROM expression TO expression DO statements0 END
     { For($2, $4, $6, $8) }
 | WHILE expression DO statements0 END { While($2, $4) }

statement:
   statement_d { $1, (extent ()) }

statements:
   statement { [$1] }
 | statement SEMICOLON statements { $1::$3 }
 | statement statements
     {
       Error.syntax_error "missing semicolon after" (item_extent 1); 
       $1::$2
     }

statements0:
		{ [] }
 | statements   { $1 }

block:
   DECLARE declarations0 BEGINBLOCK statements0 END
     { { declarations = $2 ; code = $4 } }
/* sucre syntaxique: pas oblig de mettre un DECLARE vide */
 | BEGINBLOCK statements0 END
     { { declarations = [] ; code = $2 } }

declaration:
   Identifier COLON type_spec { VariableDecl($1, $3) }
 | Identifier COLON ARRAY LBRACKET bounds RBRACKET OF type_spec
     { ArrayDecl({ array_id = $1;
         array_bounds = $5;
         array_type = $8 }) }
 | Identifier IS expression COLON type_spec { IdentityDecl($1, $3, $5) }
 | procedure { ProcedureDecl($1) }

/* Exemple de rcupration */
/* cette clause faire planter le parser sur parser_bug.lax
*/
 | error
     {
       Error.syntax_error "parse error in declaration" (extent ());
       ErrorDecl
     }

bounds:
   bound { [$1] }
 | bound COMMA bounds { $1::$3 }

bound:
   expression COLON expression { $1,$3 }

declarations:
   declaration { [$1] }
 | declaration SEMICOLON declarations { $1::$3 }
 | error SEMICOLON declarations
     {
       Error.syntax_error "parse error in declaration" (extent ());
       ErrorDecl::$3
     }

declarations0:
		{ [] }
 | declarations { $1 }

procedure:
   PROCEDURE Identifier LPAREN parameters RPAREN result_type SEMICOLON expression
       {{ proc_name = $2;
          proc_args = $4;
          proc_type = $6;
          proc_value = Declared $8 }}

 | PROCEDURE Identifier LPAREN type_parameters RPAREN result_type EQUAL Identifier
       {{ proc_name = $2;
          proc_args = $4;
          proc_type = $6;
          proc_value = External $8 }}

 | PROCEDURE Identifier result_type SEMICOLON expression
       {{ proc_name = $2;
          proc_args = [];
          proc_type = $3;
          proc_value = Declared $5 }}

 | PROCEDURE Identifier result_type EQUAL Identifier
       {{ proc_name = $2;
          proc_args = [];
          proc_type = $3;
          proc_value = External $5 }}

/*
 | PROCEDURE Identifier LPAREN error RPAREN result_type SEMICOLON expression
       {  Error.syntax_error "parse error in parameters" (item_extent 4);
          { proc_name = $2;
            proc_args = [];
            proc_type = $6;
            proc_value = $8 }}
*/

parameter:
   Identifier COLON type_spec { ($1, $3) }

parameters:
   parameter { [$1] }
   | parameter SEMICOLON parameters { $1::$3 }
   | error SEMICOLON parameters
     {
       Error.syntax_error "parse error in parameters, skipping" (extent ());
       $3
     }

type_parameters:
   type_spec { [("unknown", $1)] }
   | type_spec SEMICOLON type_parameters { ("ungnown", $1)::$3 }
   | error SEMICOLON type_parameters
     {
       Error.syntax_error "parse error in parameters of external, skipping" (extent ());
       $3
     }

result_type:
	{ No_result_type }
 | COLON type_spec	{ Result_type $2 }

type_spec:
   Identifier	{ NamedType $1 }
 | REF type_spec { RefType $2 }
 | REF ARRAY LBRACKET dimensions RBRACKET OF type_spec { RefArrayType($7, $4) }
 | PROCEDURE result_type { ProcType([], $2) }
 | PROCEDURE LPAREN type_specs RPAREN result_type { ProcType($3, $5) }
 | error
     {
       Error.syntax_error "parse error in type" (extent ());
       ErrorType
     }

/* Les types des tableaux ne contiennent que le nombre de dimensions,
   donn par un nombre de virgules. Par exemple, le type
   array[,,] of float est le type des tableaux de flottants  3 dimensions. */
dimensions:
   			{ 1 }
 | COMMA dimensions 	{ succ $2 }

type_specs:
   type_spec { [$1] }
 | type_spec SEMICOLON type_specs { $1::$3 }

