(*********************************************************
Gnration de code Sparc
(pas de placement de variables en registres)


$Id: sparcgen.ml,v 1.1.1.1 1998/06/23 22:35:09 monniaux Exp $
*********************************************************)

open Sparc;;

open Typeintermed;;

(* Allocation des registres :
on attribue les 20 registres de %o7  %i1, sauf %sp, %fp et %i7,
en commenant par les feuilles.
La raison de cet ordre est qu'un appel de procdure n'a alors qu'
dire qu'il occupe 7 registres (%o0  %o6).
Lorsqu'il n'y a plus de registre, on sauve dans des temporaires.
%g1 sera rserv comme "scratch pad".

Le (si : stack_state) reprsente le nombre de registres (commenant  0)
dans lesquels il ne faut pas mettre la valeur de retour.

La valeur stack_state retourne indique le nombre de registres
(commenant  0) crass.
*)

(* Dtail piquant : les numros admissibles pour Txx sont dans dans
   /usr/include/sys/trap.h
*)

type stack_state = int;;

let fregister_id = function
  x when x<0 -> failwith "negative offset in float register stack"
| x when x<15 -> 2*x
| _ -> failwith "float register stack overflow";;

let register_id = function
  x when x<0 -> failwith "negative offset in integer register stack"
| 0 -> 31
| x when x<15 -> 30-x
| x when x<20 -> 28-x
| _ -> failwith "integer register stack overflow";;

(*{{{*)
let nb_hardware_registers = 20
and overflow_ret_register = 8
and return_register = 24
and nb_hardware_fregisters = 15
and overflow_ret_fregister = 30
and return_fregister = 0
and input_registers = 8
and save_old_frame_offset = -8
and announce_integer_stack = 100000
and announce_float_stack = 200000
and bottom_of_frame_size = 24*4;;
(*}}}*)

let rec gen_args ofs =
(*{{{*)
  function
    [] -> (Dlist.singleton (Opcode Nop)), 7, 15
  | h::t -> let block2, s2, sf2, r2 =
              match h with
                Pexpr(_, p, _) -> ptr_eval p 0
              | Iexpr(i) -> int_eval i 0
            and block1, s1, sf1 = gen_args (ofs+4) t
            in  Dlist.paste_before block2 block1;
 Dlist.insert_before (Opcode(Store(Word, r2, stack_pointer, (Imm13 ofs)))) block1;
                block1, (max s1 s2), (max sf1 sf2)

and create_call (ptr: pexpr) (args: expr list) =
  let nb_args = List.length args
  in  let block1, s1, sf1 = gen_args parameters_offset args 
      in  let s1, sf1 = ( match ptr with
              AsmLabel(l) ->
                for i=0 to pred (min 6 nb_args)
                do
                  Dlist.insert_before
  (Opcode(Load(Word, stack_pointer, Imm13(parameters_offset+4*i),
        return_register + i))) block1
                done;
                Dlist.insert_before (Opcode(Call l)) block1;
                s1, sf1
            | _ -> let block2, s2, sf2, r2 = ptr_eval ptr 7
                   in  Dlist.paste_before block2 block1;
                       for i=0 to pred (min 6 nb_args)
                       do
                         Dlist.insert_before
  (Opcode(Load(Word, stack_pointer, Imm13(parameters_offset+4*i),
        return_register + i))) block1
                       done;
                       Dlist.insert_before (Opcode(Callr r2)) block1;
                       (max s1 s2), (max sf1 sf2)
          ) in
          Dlist.insert_before (Opcode Nop) block1;
          let shift = ((nb_args-6)*4+7) land -8
          in if shift>0
             then ( Dlist.insert_before (Opcode(
  Triadic(Add, false, stack_pointer, (Imm13 shift), stack_pointer))) block1;
                    (Dlist.insert_before (Opcode(
  Triadic(Sub, false, stack_pointer, (Imm13 shift), stack_pointer))) block1),
                    s1, sf1
                  )
             else block1, s1, sf1
(*}}}*)

and branch_if (e : bexpr) (target: Sparc.label) =
(*{{{*)
  match e with
    BNot(BNot(x)) -> branch_if x target

  | True -> (create_branch Always target), 0, 0
  | BNot(True) -> branch_if False target

  | False -> (Dlist.singleton (Opcode(Nop))), 0, 0
  | BNot(False) -> branch_if True target

  | BOr(x,y) -> let block1, s1, sf1 = branch_if x target
                and block2, s2, sf2 = branch_if y target
                in  Dlist.paste_before block2 block1;
                    block1, (max s1 s2), (max sf1 sf2)
  | BNot(BOr(x,y)) -> branch_if (BAnd((BNot(x)),(BNot(y)))) target

  | BAnd(x,y) -> let label = make_label()
                 in  let block1, s1, sf1 = branch_if (BNot x) label
                     and block2, s2, sf2 = branch_if y target
                     in  Dlist.paste_before block2 block1;
                         Dlist.insert_before (Label(label)) block1;
                         block1, (max s1 s2), (max sf1 sf2)
  | BNot(BAnd(x,y)) -> branch_if (BOr((BNot(x)),(BNot(y)))) target

  | BEgal(a, b) -> let label1 = make_label() and label2 = make_label()
                   in  let block1, s1, sf1 = branch_if a label1
                       and block2, s2, sf2 = branch_if (BNot b) target
                       and block3, s3, sf3 = branch_if b target
                       in  Dlist.paste_before block2 block1;
                           Dlist.paste_before (create_branch Always label2) block1;
                           Dlist.insert_before (Label label1) block1;
                           Dlist.paste_before block3 block1;
                           Dlist.insert_before (Label label2) block1;
                           block1, (max (max s1 s2) s3), (max (max sf1 sf2) sf3)
  | BNot(BEgal(a, b)) -> branch_if (BEgal((BNot a),b)) target

  | IInf(a, b) -> int_compare Less target a b
  | BNot(IInf(a, b)) -> int_compare GreaterOrEqual target a b

  | ISup(a, b) -> int_compare Greater target a b
  | BNot(ISup(a, b)) -> int_compare LessOrEqual target a b

  | IEgal(a, b) -> int_compare Equal target a b 
  | BNot(IEgal(a, b)) -> int_compare NotEqual target a b 

  | PEgal(a, b) -> ptr_compare Equal target a b 
  | BNot(PEgal(a, b)) -> ptr_compare NotEqual target a b 

  | FInf(a, b) -> float_compare F_Less target a b
  | BNot(FInf(a, b)) -> float_compare F_GreaterOrEqual target a b

  | FSup(a, b) -> float_compare F_Greater target a b
  | BNot(FSup(a, b)) -> float_compare F_LessOrEqual target a b

  | FEgal(a, b) -> float_compare F_Equal target a b 
  | BNot(FEgal(a, b)) -> float_compare F_NotEqual target a b 

  | Bseq(a, b) -> void_eval a; branch_if b target
  | BNot(Bseq(a, b)) -> branch_if (Bseq(a, (BNot b))) target

  | B_ITE(a, b, c) -> let label = make_label() and label2=make_label()
                     in let block1, s1, sf1 = branch_if (BNot a) label
                        and block2, s2, sf2 = branch_if b target
                        and block3, s3, sf3 = branch_if c target
                        in  Dlist.paste_before block2 block1;
                            Dlist.paste_before (create_branch Always label2) block1;
                            Dlist.insert_before (Label label) block1;
                            Dlist.paste_before block3 block1;
                            Dlist.insert_before (Label label2) block1;
                            block1, (max s1 (max s2 s3)), (max sf1 (max sf2 sf3))
  | BNot(B_ITE(a, b, c)) -> branch_if (B_ITE(a, (BNot b), (BNot c))) target

  | Binvseq(a, b) -> let label = make_label()
                     in let block1, s1, sf1 = branch_if (BNot a) label
                        and block2, s2, sf2 = void_eval b
                        and block3, s3, sf3 = void_eval b
                        in  Dlist.paste_before block2 block1;
                            Dlist.paste_before (create_branch Always target) block1;
                            Dlist.insert_before (Label label) block1;
                            Dlist.paste_before block3 block1;
                            block1, (max s1 (max s2 s3)), (max sf1 (max sf2 sf3))

  | BNot(Binvseq(a, b)) -> branch_if (Binvseq((BNot a), b)) target

  | Bprocedure(ptr, arg) ->
      branch_if (BNot(IEgal((Iprocedure(ptr, arg)),(IConst 0)))) target
  | BNot(Bprocedure(ptr, arg)) ->
      branch_if (IEgal((Iprocedure(ptr, arg)),(IConst 0))) target

  | BBlock(_) | BNot(BBlock(_)) -> failwith "BBlock ??"
  | B_peek(ptr) -> branch_if (BNot(IEgal((I_peek ptr),(IConst 0)))) target
  | BNot(B_peek(ptr)) -> branch_if (IEgal((I_peek ptr),(IConst 0))) target
(*}}}*)

and int_compare (typ : condition) (target: label) (a: iexpr) (b: iexpr) =
(*{{{*)
      match b with
        IConst(c) when c>= -4096 & c<4096 ->
          let block1, s1, sf1, r1 = int_eval a 0
          in  Dlist.insert_before
                (Opcode(Triadic(Sub, true, r1, Imm13(c), 0))) block1;
              Dlist.paste_before (create_branch typ target) block1;
              block1, s1, sf1
      | _ -> let block2, s2, sf2, r2 = int_eval b 0
             in  let block1, s1, sf1, r1 = int_eval a s2
                 in  Dlist.paste_before block2 block1;
                     Dlist.insert_before
                       (Opcode(Triadic(Sub, true, r1 , Reg(r2), 0)))
                     block1;
            Dlist.paste_before (create_branch typ target) block1;
            block1, (max s1 s2), (max sf1 sf2)
(*}}}*)

and ptr_compare (typ : condition) (target: label) (a: pexpr) (b: pexpr) =
(*{{{*)
           let block2, s2, sf2, r2 = ptr_eval b 0
             in  let block1, s1, sf1, r1 = ptr_eval a s2
                 in  Dlist.paste_before block2 block1;
                     Dlist.insert_before
                       (Opcode(Triadic(Sub, true, r1 , Reg(r2), 0)))
                     block1;
            Dlist.paste_before (create_branch typ target) block1;
            block1, (max s1 s2), (max sf1 sf2)
(*}}}*)

and float_compare (typ : f_condition) (target: label) (a: fexpr) (b: fexpr) =
(*{{{*)
  let block2, s2, sf2, r2 = float_eval b 0
  in  let block1, s1, sf1, r1 = float_eval a sf2
      in  Dlist.paste_before block2 block1;
          Dlist.insert_before (Opcode(F_Cmp(r1 , r2))) block1;
          Dlist.insert_before (Opcode(Nop)) block1;
          Dlist.paste_before (create_fbranch typ target) block1;
          block1, (max s1 s2), (max sf1 sf2)
(*}}}*)

and int_eval (ex : iexpr) (si: stack_state) =
(*{{{*)
  match ex with
    IConst(0) -> (Dlist.singleton (Opcode Nop)), 0, 0, 0
  | IConst(x) -> if   si<nb_hardware_registers
                 then let r=register_id si
                      in  (create_load_const x r), (succ si), 0, r
                 else
   (create_load_const x overflow_ret_register), 0, 0, overflow_ret_register

  | I_peek(PAdd(FramePtr, IConst(ofs))) ->
      if   si<nb_hardware_registers
                 then let r=register_id si
                      in 
  (Dlist.singleton(Opcode(Load(Word, frame_pointer, Imm13(ofs), r)))),
    (succ si), 0, r
                 else
  (Dlist.singleton(Opcode(Load(Word, frame_pointer, Imm13(ofs), overflow_ret_register)))),
    0, 0, overflow_ret_register

  | I_peek(PAdd(x, IConst(ofs))) ->
      let block1, s1, sf1, r1 = ptr_eval x 0
      in
      if   si<nb_hardware_registers
                 then ( let r=register_id si
                      in 
  (Dlist.insert_before (Opcode(Load(Word, r1, Imm13(ofs), r))) block1);
    block1, (max s1 (succ si)), sf1, r )
                 else
 (  (Dlist.insert_before (Opcode(Load(Word, r1, Imm13(ofs),
     overflow_ret_register))) block1);
    block1, s1, sf1, overflow_ret_register )

  | I_peek(x) -> let block1, s1, sf1, r1 = ptr_eval x si
                 in  Dlist.insert_before (Opcode(Load(Word, r1, Reg(0), r1))) block1;
                     block1, s1, sf1, r1

  | INeg(x) -> let block1, s1, sf1, r1 = int_eval x si
                 in  Dlist.insert_before
      (Opcode(Triadic(Sub, false, 0, Reg(r1), r1))) block1;
                     block1, s1, sf1, r1

  | IAdd(x, y) -> create_triadic Add_op x y si
  | ISub(x, y) -> create_triadic Sub_op x y si
  | IMul(x, y) -> create_triadic Mul_op x y si
  | IDiv(x, y) -> create_triadic Div_op x y si
  | IMod(x, y) -> create_triadic Mod_op x y si

  | I_ITE(t, x, y) ->
      let label1 = make_label() and label2 = make_label()
      in  let block1, s1, sf1, r1 = int_eval x si
          and block2, s2, sf2, r2 = int_eval y si
          and block0, s0, sf0 = branch_if t label1
          in  if r1<>0 then
              (
              Dlist.paste_before block2 block0;
              (if r1 <> r2 then (
Dlist.insert_before (Opcode(Triadic(Add, false, r2, Reg(0), r1))) block0; ()));
              Dlist.paste_before (create_branch Always label2) block0;
              Dlist.insert_before (Label(label1)) block0;
              Dlist.paste_before block1 block0;
              Dlist.insert_before (Label(label2)) block0;
              block0, (max (max s1 s0) s2), (max (max sf1 sf0) sf2), r1
              ) else
              (
              Dlist.paste_before block2 block0;
              Dlist.paste_before (create_branch Always label2) block0;
              Dlist.insert_before (Label(label1)) block0;
              Dlist.paste_before block1 block0;
              (if r1 <> r2 then (
Dlist.insert_before (Opcode(Triadic(Add, false, r1, Reg(0), r2))) block0; ()));
              Dlist.insert_before (Label(label2)) block0;
              block0, (max (max s1 s0) s2), (max (max sf1 sf0) sf2), r2
	      )

  | Iseq(e1, e2)    -> let block1, s1, sf1 = void_eval e1
                       and block2, s2, sf2, r2 = int_eval e2 si
                       in  Dlist.paste_before block2 block1;
              block1, (max s1 s2), (max sf1 sf2), r2
  | Iinvseq(e1, e2) -> let block2, s2, sf2 = void_eval e2
                       in  let block1, s1, sf1, r1 = int_eval e1 (max si s2)
                       in  Dlist.paste_before block2 block1;
                           block1, (max s1 s2), (max sf1 sf2), r1
  | Iprocedure(proc, args) -> let a, b, c = (create_call proc args)
                              in  a, b, c, return_register

  | IBlock(_) -> failwith "IBlock ??"
(*}}}*)

and ptr_eval (ex : pexpr) (si: stack_state) =
(*{{{*)
  match ex with
    Nil -> (Dlist.singleton (Opcode Nop)), 0, 0, 0
  | AsmLabel(x) ->if   si<nb_hardware_registers
                 then let r=register_id si
                      in  (create_load_label x r), (succ si), 0, r
                 else
   (create_load_label x overflow_ret_register), 0, 0, overflow_ret_register

  | FramePtr -> if   si<nb_hardware_registers
                 then let r=register_id si
                 in
 (Dlist.singleton(Opcode(Triadic(Add, false, frame_pointer, Reg(0), r)))),
                      (succ si), 0, r
                 else
 (Dlist.singleton(Opcode(Triadic(Add, false, frame_pointer, Reg(0),
  overflow_ret_register)))),
                      0, 0, overflow_ret_register
  | PAdd(x, y) -> create_ptriadic Add_op x y si
  | P_ITE(t, x, y) ->
      let label1 = make_label() and label2 = make_label()
      in  let block1, s1, sf1, r1 = ptr_eval x si
          and block2, s2, sf2, r2 = ptr_eval y si
          and block0, s0, sf0 = branch_if t label1
          in  if r1<>0 then
              (
              Dlist.paste_before block2 block0;
              (if r1 <> r2 then (
Dlist.insert_before (Opcode(Triadic(Add, false, r2, Reg(0), r1))) block0; ()));
              Dlist.paste_before (create_branch Always label2) block0;
              Dlist.insert_before (Label(label1)) block0;
              Dlist.paste_before block1 block0;
              Dlist.insert_before (Label(label2)) block0;
              block0, (max (max s1 s0) s2), (max (max sf1 sf0) sf2), r1
              ) else
              (
              Dlist.paste_before block2 block0;
              Dlist.paste_before (create_branch Always label2) block0;
              Dlist.insert_before (Label(label1)) block0;
              Dlist.paste_before block1 block0;
              (if r1 <> r2 then (
Dlist.insert_before (Opcode(Triadic(Add, false, r1, Reg(0), r2))) block0; ()));
              Dlist.insert_before (Label(label2)) block0;
              block0, (max (max s1 s0) s2), (max (max sf1 sf0) sf2), r2
	      )

  | Pseq(e1, e2)    -> let block1, s1, sf1 = void_eval e1
                        and block2, s2, sf2, r2 = ptr_eval e2 si
                        in  Dlist.paste_before block2 block1;
               block1, (max s1 s2), (max sf1 sf2), r2

  | P_peek(PAdd(FramePtr, IConst(ofs))) ->
      if   si<nb_hardware_registers
                 then let r=register_id si
                      in 
  (Dlist.singleton(Opcode(Load(Word, frame_pointer, Imm13(ofs), r)))),
    (succ si), 0, r
                 else
  (Dlist.singleton(Opcode(Load(Word, frame_pointer, Imm13(ofs), overflow_ret_register)))),
    0, 0, overflow_ret_register

  | P_peek(PAdd(x, IConst(ofs))) ->
      let block1, s1, sf1, r1 = ptr_eval x 0
      in
      if   si<nb_hardware_registers
                 then ( let r=register_id si
                      in 
  (Dlist.insert_before (Opcode(Load(Word, r1, Imm13(ofs), r))) block1);
    block1, (max s1 (succ si)), sf1, r )
                 else
 (  (Dlist.insert_before (Opcode(Load(Word, r1, Imm13(ofs),
     overflow_ret_register))) block1);
    block1, s1, sf1, overflow_ret_register )

  | P_peek(AsmLabel l) ->
      let block1 =
     Dlist.singleton(Opcode(Sethi((Hi_l l), scratch_register)))
      in if si<nb_hardware_registers
                         then ( let r = register_id si
                                in  Dlist.insert_before
    (Opcode(Load(Word, scratch_register, (Lo_l l), r))) block1;
                       block1, (succ si), 0, r
                              )
                         else ( Dlist.insert_before
 (Opcode(Load(Word, scratch_register, (Lo_l l), overflow_ret_register)))
    block1;
                   block1, 0, 0, overflow_ret_register
                              )

  | P_peek(x) -> let block1, s1, sf1, r1 = ptr_eval x si
                 in  Dlist.insert_before (Opcode(Load(Word, r1, Reg(0), r1))) block1;
                     block1, s1, sf1, r1

  | FAssignment(p, v) -> let block2, s2, sf2, r2 = ptr_eval p si
                         and block1, s1, sf1, r1 = float_eval v 0
                         in  Dlist.paste_before block2 block1;
          Dlist.insert_before (Opcode(F_Store(Double, r1, r2, Reg(0)))) block1;
                             block1, (max s1 s2), (max sf1 sf2), r2
   
  | IAssignment(p, v) -> let block2, s2, sf2, r2 = ptr_eval p si
             in  let block1, s1, sf1, r1 = int_eval v s2
                 in  let r1,s2 =
                       ( if   s2<nb_hardware_registers
                         then ( Dlist.paste_before block2 block1;
                                r1, s2
                              )
                         else ( Dlist.insert_before
  (Opcode(Store(Word, r1, frame_pointer, Imm13(announce_integer_stack - nb_hardware_registers + s2)))) block1;
                                Dlist.paste_before block2 block1;
                                Dlist.insert_before
  (Opcode(Load(Word, frame_pointer, Imm13(announce_integer_stack - nb_hardware_registers + s2), overflow_ret_register))) block1;
                                overflow_ret_register, (succ s2)
                              )
                       )
                     in 
               Dlist.insert_before (Opcode(Store(Word, r1, r2, (Reg 0)))) block1;
               block1, (max s1 s2), (max sf1 sf2), r2

  | PAssignment(p, v) -> let block2, s2, sf2, r2 = ptr_eval p si
             in  let block1, s1, sf1, r1 = ptr_eval v s2
                 in  let r1,s2 =
                       ( if   s2<nb_hardware_registers
                         then ( Dlist.paste_before block2 block1;
                                r1, s2
                              )
                         else ( Dlist.insert_before
  (Opcode(Store(Word, r1, frame_pointer, Imm13(announce_integer_stack - nb_hardware_registers + s2)))) block1;
                                Dlist.paste_before block2 block1;
                                Dlist.insert_before
  (Opcode(Load(Word, frame_pointer, Imm13(announce_integer_stack - nb_hardware_registers + s2), overflow_ret_register))) block1;
                                overflow_ret_register, (succ s2)
                              )
                       )
                     in 
               Dlist.insert_before (Opcode(Store(Word, r1, r2, (Reg 0)))) block1;
               block1, (max s1 s2), (max sf1 sf2), r2
   | Alloca(s) ->
       let block1, s1, sf1, r1 = int_eval s si
       in  Dlist.insert_before
  (Opcode(Triadic(Add, false, stack_pointer, Reg(0), scratch_register))) block1;
           Dlist.insert_before
  (Opcode(Triadic(Add, false, r1, Imm13(15), r1))) block1;
           Dlist.insert_before
  (Opcode(Triadic(Sparc.And, false, r1, Imm13(-8), r1))) block1;
           Dlist.insert_before
  (Opcode(Triadic(Sub, false, stack_pointer, Reg(r1), stack_pointer))) block1;
           Dlist.insert_before (Opcode(
 Store(Word, scratch_register, stack_pointer, Imm13(bottom_of_frame_size)))) block1;
           Dlist.insert_before
  (Opcode(Triadic(Add, false, stack_pointer, Imm13(8+bottom_of_frame_size), r1))) block1;
           block1, s1, sf1, r1

  | Pinvseq(e1, e2) -> let block2, s2, sf2 = void_eval e2
                       in  let block1, s1, sf1, r1 = ptr_eval e1 (max si s2)
                       in  Dlist.paste_before block2 block1;
                           block1, (max s1 s2), (max sf1 sf2), r1

  | Pprocedure(proc, args) -> let a, b, c = (create_call proc args)
                              in  a, b, c, return_register

  | BAssignment(ptr, expr) ->
      ptr_eval (IAssignment(ptr, (bool_to_int expr))) 0

  | Var(_) -> failwith "Var ??"
  | PBlock(_) -> failwith "PBlock ??"
(*}}}*)

and bool_to_int expr = I_ITE(expr, (IConst 1), (IConst 0))

and create_triadic (typ : binary_op) (x: iexpr) (y: iexpr) (si: stack_state) =
(*{{{*)
  match y with
        IConst(c) when c>= -4096 & c<4096
          -> let block1, s1, sf1, r1 = int_eval x 0
             in  if   si<nb_hardware_registers
                 then ( let r = register_id si
                        in  Dlist.paste_before
                      ( create_triadic_op typ r1 (Imm13 c) r) block1;
                            block1, (max s1 (succ si)), sf1, r
                      )
                 else ( Dlist.paste_before
    (create_triadic_op typ r1 (Imm13 c) overflow_ret_register) block1;
                        block1, s1, sf1, overflow_ret_register
                      )

      | _ -> let block2, s2, sf2, r2 = int_eval y 0
             in  let block1, s1, sf1, r1 = int_eval x s2
                 in  let r1,s2 =
                       ( if   s2<nb_hardware_registers
                         then ( Dlist.paste_before block2 block1;
                                r1, s2
                              )
                         else ( Dlist.insert_before
  (Opcode(Store(Word, r1, frame_pointer, Imm13(announce_integer_stack - nb_hardware_registers + s2)))) block1;
                                Dlist.paste_before block2 block1;
                                Dlist.insert_before
  (Opcode(Load(Word, frame_pointer, Imm13(announce_integer_stack - nb_hardware_registers + s2), overflow_ret_register))) block1;
                                overflow_ret_register, (succ s2)
                              )
                       )
                     in  if si<nb_hardware_registers
                         then ( let r = register_id si
                                in  Dlist.paste_before
                   (create_triadic_op typ r1 (Reg r2) r) block1;
                       block1, (max (max s1 s2) (succ si)), (max sf1 sf2), r
                              )
                         else ( Dlist.paste_before
      (create_triadic_op typ r1 (Reg r2) overflow_ret_register) block1;
                   block1, (max s1 s2), (max sf1 sf2), overflow_ret_register
(*}}}*)
                              )
and create_ptriadic (typ : binary_op) (x: pexpr) (y: iexpr) (si: stack_state) =
(*{{{*)
  match y with
        IConst(c) when c>= -4096 & c<4096
          -> let block1, s1, sf1, r1 = ptr_eval x 0
             in  if   si<nb_hardware_registers
                 then ( let r = register_id si
                        in  Dlist.paste_before
                      ( create_triadic_op typ r1 (Imm13 c) r) block1;
                            block1, (max s1 (succ si)), sf1, r
                      )
                 else ( Dlist.paste_before
    (create_triadic_op typ r1 (Imm13 c) overflow_ret_register) block1;
                        block1, s1, sf1, overflow_ret_register
                      )

      | _ -> let block2, s2, sf2, r2 = int_eval y 0
             in  let block1, s1, sf1, r1 = ptr_eval x s2
                 in  let r1,s2 =
                       ( if   s2<nb_hardware_registers
                         then ( Dlist.paste_before block2 block1;
                                r1, s2
                              )
                         else ( Dlist.insert_before
  (Opcode(Store(Word, r1, frame_pointer, Imm13(announce_integer_stack - nb_hardware_registers + s2)))) block1;
                                Dlist.paste_before block2 block1;
                                Dlist.insert_before
  (Opcode(Load(Word, frame_pointer, Imm13(announce_integer_stack - nb_hardware_registers + s2), overflow_ret_register))) block1;
                                overflow_ret_register, (succ s2)
                              )
                       )
                     in  if si<nb_hardware_registers
                         then ( let r = register_id si
                                in  Dlist.paste_before
                   (create_triadic_op typ r1 (Reg r2) r) block1;
                       block1, (max (max s1 s2) (succ si)), (max sf1 sf2), r
                              )
                         else ( Dlist.paste_before
      (create_triadic_op typ r1 (Reg r2) overflow_ret_register) block1;
                   block1, (max s1 s2), (max sf1 sf2), overflow_ret_register
(*}}}*)
                              )
and float_eval (ex : fexpr) (si: stack_state) =
(*{{{*)
  match ex with
    F_from_i(x) -> let block1, s1, sf1, r1 = int_eval x 0
                   in if   si<nb_hardware_fregisters
                      then ( let r = fregister_id si
                             in  Dlist.paste_before (create_dtoi r1 r) block1;
                                 block1, s1, (max sf1 (succ si)), r
                           )
                      else (
       Dlist.paste_before (create_dtoi r1 overflow_ret_fregister) block1;
                             block1, s1, sf1, overflow_ret_fregister
                           )

  | FNeg(x) -> let block1, s1, sf1, r1 = float_eval x si
               in  Dlist.insert_before(Opcode(F_Diadic(F_Neg, r1, r1))) block1;
                   block1, s1, sf1, r1
  | FAdd(x, y) -> create_ftriadic F_Add x y si
  | FSub(x, y) -> create_ftriadic F_Sub x y si
  | FMul(x, y) -> create_ftriadic F_Mul x y si
  | FDiv(x, y) -> create_ftriadic F_Div x y si
  | F_ITE(t, x, y) ->
      let label1 = make_label() and label2 = make_label()
      in  let block1, s1, sf1, r1 = float_eval x si
          and block2, s2, sf2, r2 = float_eval y si
          and block0, s0, sf0 = branch_if t label1
          in  Dlist.paste_before block2 block0;
              if r1 <> r2 then
                ( Dlist.paste_before (create_fmove r2 r1) block0; () );
              Dlist.paste_before (create_branch Always label2) block0;
              Dlist.insert_before (Label(label1)) block0;
              Dlist.paste_before block1 block0;
              Dlist.insert_before (Label(label2)) block0;
              block0, (max (max s1 s0) s2), (max (max sf1 sf0) sf2), r2

  | F_peek(AsmLabel(l)) ->
      let block1 =
     Dlist.singleton(Opcode(Sethi((Hi_l l), scratch_register)))
      in if si<nb_hardware_fregisters
                         then ( let r = fregister_id si
                                in  Dlist.insert_before
    (Opcode(F_Load(Double, scratch_register, (Lo_l l), r))) block1;
                       block1, 0, (succ si), r
                              )
                         else ( Dlist.insert_before
 (Opcode(F_Load(Double, scratch_register, (Lo_l l), overflow_ret_fregister)))
    block1;
                   block1, 0, 0, overflow_ret_fregister
                              )
  | F_peek(PAdd(FramePtr, IConst(ofs))) ->
      if   si<nb_hardware_fregisters
                 then let r=fregister_id si
                      in 
  (Dlist.singleton(Opcode(F_Load(Double, frame_pointer, Imm13(ofs), r)))),
    0, (succ si), r
                 else
  (Dlist.singleton(Opcode(F_Load(Double, frame_pointer, Imm13(ofs), overflow_ret_register)))),
    0, 0, overflow_ret_fregister

  | F_peek(PAdd(x, IConst(ofs))) ->
      let block1, s1, sf1, r1 = ptr_eval x 0
      in
      if   si<nb_hardware_fregisters
                 then ( let r=fregister_id si
                      in 
  (Dlist.insert_before (Opcode(F_Load(Double, r1, Imm13(ofs), r))) block1);
    block1, s1, (max sf1 (succ si)), r )
                 else
 (  (Dlist.insert_before (Opcode(F_Load(Double, r1, Imm13(ofs),
     overflow_ret_fregister))) block1);
    block1, s1, sf1, overflow_ret_fregister )

  | F_peek(p) -> let block1, s1, sf1, r1 = ptr_eval p 0
                 in  if si<nb_hardware_fregisters
                         then ( let r = fregister_id si
                                in  Dlist.insert_before
                   (Opcode(F_Load(Double, r1, (Reg 0), r))) block1;
                       block1, s1, (max (succ si) sf1), r
                              )
                         else ( Dlist.insert_before
         (Opcode(F_Load(Double, r1, (Reg 0), overflow_ret_fregister))) block1;
                   block1, s1, sf1, overflow_ret_fregister
                              )

  | Fseq(e1, e2)    -> let block1, s1, sf1 = void_eval e1
                       and block2, s2, sf2, r2 = float_eval e2 si
                       in  Dlist.paste_before block2 block1;
                           block1, (max s1 s2), (max sf1 sf2), r2

  | FConst(x) -> float_eval (F_peek(AsmLabel (gen_fconst_label (hash_into float_consts x float_consts_index)))) si
  | Finvseq(e1, e2) -> let block2, s2, sf2 = void_eval e2
                       in  let block1, s1, sf1, r1 = float_eval e1 (max si sf2)
                       in  Dlist.paste_before block2 block1;
                           block1, (max s1 s2), (max sf1 sf2), r1

  | Fprocedure(proc, args) -> let a, b, c = (create_call proc args)
                              in  a, b, c, return_fregister

  | FBlock(_) -> failwith "FBlock ??"
(*}}}*)

and create_ftriadic (typ : f_triadic) (x: fexpr) (y: fexpr) (si: stack_state) =
(*{{{*)
  let block2, s2, sf2, r2 = float_eval y 0
  in  let block1, s1, sf1, r1 = float_eval x sf2
      in  let r1,sf2 =
                       ( if   sf2<nb_hardware_fregisters
                         then ( Dlist.paste_before block2 block1;
                                r1, sf2
                              )
                         else ( Dlist.insert_before
  (Opcode(F_Store(Double, r1, frame_pointer, Imm13(announce_float_stack - nb_hardware_fregisters + sf2)))) block1;
                                Dlist.paste_before block2 block1;
                                Dlist.insert_before
  (Opcode(F_Load(Double, frame_pointer, Imm13(announce_float_stack - nb_hardware_fregisters + sf2), overflow_ret_fregister))) block1;
                                overflow_ret_fregister, (succ sf2)
                              )
                       )
                     in  if si<nb_hardware_fregisters
                         then ( let r = fregister_id si
                                in  Dlist.paste_before
                   (create_ftriadic_op typ r1 r2 r) block1;
                       block1, (max s1 s2), (max (succ si) (max sf1 sf2)), r
                              )
                         else ( Dlist.paste_before
      (create_ftriadic_op typ r1 r2 overflow_ret_fregister) block1;
                   block1, (max s1 s2), (max sf1 sf2), overflow_ret_fregister
(*}}}*)
                              )
and void_eval =
(*{{{*)
  function
    V_from_p(IAssignment(PAdd(FramePtr, (IConst ofs)), v)) ->
      let block1, s1, sf1, r1 = int_eval v 0
      in
Dlist.insert_before (Opcode(Store(Word, r1, frame_pointer, (Imm13 ofs)))) block1;
          block1, s1, sf1

  | V_from_p(PAssignment(PAdd(FramePtr, (IConst ofs)), v)) ->
      let block1, s1, sf1, r1 = ptr_eval v 0
      in
Dlist.insert_before (Opcode(Store(Word, r1, frame_pointer, (Imm13 ofs)))) block1;
          block1, s1, sf1

  | V_from_p(FAssignment(PAdd(FramePtr, (IConst ofs)), v)) ->
      let block1, s1, sf1, r1 = float_eval v 0
      in
Dlist.insert_before(Opcode(F_Store(Double, r1, frame_pointer, (Imm13 ofs)))) block1;
          block1, s1, sf1

  | V_from_i(iexpr) -> let block1, s1, sf1, _ = int_eval iexpr 0
                       in  block1, s1, sf1
  | V_from_p(pexpr) -> let block1, s1, sf1, _ = ptr_eval pexpr 0
                       in  block1, s1, sf1
  | V_from_f(fexpr) -> let block1, s1, sf1, _ = float_eval fexpr 0
                       in  block1, s1, sf1
  | V_from_b(bexpr) -> let label=make_label()
                       in  let block1, s1, sf1 = branch_if bexpr label
                           in  Dlist.insert_before (Label label) block1;
                               block1, s1, sf1
  | Vseq(e1, e2)    -> let block1, s1, sf1 = void_eval e1
                       and block2, s2, sf2 = void_eval e2
                       in  Dlist.paste_before block2 block1;
                           block1, (max s1 s2), (max sf1 sf2)
  | V_ITE(t, x, y) ->
      let label1 = make_label() and label2 = make_label()
      in  let block1, s1, sf1 = void_eval x
          and block2, s2, sf2 = void_eval y
          and block0, s0, sf0 = branch_if t label1
          in  Dlist.paste_before block2 block0;
              Dlist.paste_before (create_branch Always label2) block0;
              Dlist.insert_before (Label(label1)) block0;
              Dlist.paste_before block1 block0;
              Dlist.insert_before (Label(label2)) block0;
              block0, (max (max s1 s0) s2), (max (max sf1 sf0) sf2)
  | V_IT(t, x) ->
      let label1 = make_label()
      in  let block1, s1, sf1 = void_eval x
          and block0, s0, sf0 = branch_if (BNot t) label1
          in  Dlist.paste_before block1 block0;
              Dlist.insert_before (Label(label1)) block0;
              block0, (max s1 s0), (max sf1 sf0)
  | While(t, x) ->
      let label1 = make_label() and label2 = make_label()
      in  let block1, s1, sf1 = void_eval x
          and block2, s2, sf2 = branch_if t label1
          and block0 = create_branch Always label2
          in  Dlist.insert_before (Label(label1)) block0;
              Dlist.paste_before block1 block0;
              Dlist.insert_before (Label(label2)) block0;
              Dlist.paste_before block2 block0;
              block0, (max s1 s2), (max sf1 sf2)

  | For(i, a, b, block) -> failwith "for not implemented"
     (* void_eval (Vseq((V_from_p (IAssignment(i, q))),
        (While(Vseq(block,
          (V_from_p (IAssignment(i, (IAdd((I_peek i), (IConst 1)))))))))))
     *)

  | Free -> (Dlist.singleton (Opcode
   (Load(Word, stack_pointer, Imm13(bottom_of_frame_size), stack_pointer)))), 0, 0

  | Erreur -> let block1 = Dlist.singleton (Opcode(Call "lax__bad_dimensions"))
              in  Dlist.insert_before (Opcode(Nop)) block1;
                  block1, 0, 0

  | NNop -> (Dlist.singleton (Opcode(Nop))), 0, 0

  | Errorexpr -> failwith "errorexpr !?"

  | CopyBlock(src, dst, len) -> create_copyblock src dst len Word
  | FCopyBlock(src, dst, len) -> create_copyblock src dst len DoubleWord

  | Vprocedure(proc, args) -> create_call proc args

  | VBlock(_) -> failwith "VBlock ??"
(*}}}*)

and create_copyblock src dst len typ =
(*{{{*)
      let block3, s3, sf3, r3 = int_eval len 0
      in let block2, s2, sf2, r2 = ptr_eval dst s3
      in let block1, s1, sf1, r1 = ptr_eval src (max s2 s3)
      in let r2,s3 = (if s3<nb_hardware_registers
                         then ( Dlist.paste_before block3 block2;
                                r2, s3
                              )
                         else ( Dlist.insert_before
  (Opcode(Store(Word, r2, frame_pointer, Imm13(announce_integer_stack - nb_hardware_registers + s3)))) block2;
                                Dlist.paste_before block3 block2;
                                Dlist.insert_before
  (Opcode(Load(Word, frame_pointer, Imm13(announce_integer_stack - nb_hardware_registers + s3), overflow_ret_register))) block2;
                                overflow_ret_register, (succ s3)
                              ))
      in let r1,s2 =
                       ( if   s2<nb_hardware_registers
                         then ( Dlist.paste_before block2 block1;
                                r1, s2
                              )
                         else let meuh = if r1<>31 && r2<>31 then 31
                         else (if r1<>29 && r2<>29 then 29
                         else 28)
                              in
                              ( Dlist.insert_before
  (Opcode(Store(Word, r1, frame_pointer, Imm13(announce_integer_stack - nb_hardware_registers + s2)))) block1;
                                Dlist.paste_before block2 block1;
                                Dlist.insert_before
  (Opcode(Load(Word, frame_pointer, Imm13(announce_integer_stack - nb_hardware_registers + s2), meuh))) block1;
                                meuh, (succ s2)
                              )
                       )
      and label1 = make_label () and label2 = make_label()
      in Dlist.insert_before (Opcode(Triadic(Add, true, r3, Reg(0), 0))) block1;
         Dlist.paste_before (create_branch Less label2) block1;
         Dlist.insert_before (Label(label1)) block1;

         ( match typ with
             Word ->
               Dlist.insert_before (Opcode(Load(Word, r1, Reg(r3), scratch_register))) block1;
               Dlist.insert_before (Opcode(Store(Word, scratch_register, r2, Reg(r3)))) block1;
           | DoubleWord ->
               Dlist.insert_before (Opcode(F_Load(Double, r1, Reg(r3), overflow_ret_fregister))) block1;
               Dlist.insert_before (Opcode(F_Store(Double, overflow_ret_fregister, r2, Reg(r3)))) block1;
         );

         Dlist.insert_before (Opcode(Triadic(Sub, true, r3, Imm13(
           match typ with Word -> 4 | DoubleWord -> 8), r3))) block1;
         Dlist.paste_before (create_branch GreaterOrEqual label1) block1;
         Dlist.insert_before (Label(label2)) block1;
         block1, (max 3 (max s1 (max s2 s3))), (max sf1 (max sf2 sf3))
(*}}}*);;

let map_temp_stack offset f_offset sequence =
(*{{{*)
  Dlist.p_iter (fun ins ->
    match Dlist.contents ins with
      Opcode(Load(Word, frame_pointer, Imm13(z), r)) when z>=announce_integer_stack ->
        let w=offset+(z-announce_integer_stack)*4
        in  Dlist.replace ins (Opcode(Load(Word, frame_pointer, Imm13(w), r)))
    | Opcode(Store(Word, r, frame_pointer, Imm13(z))) when z>=announce_integer_stack ->
        let w=offset+(z-announce_integer_stack)*4
        in  Dlist.replace ins (Opcode(Store(Word, r, frame_pointer, Imm13(w))))

    | Opcode(F_Load(Double, frame_pointer, Imm13(z), r)) when z>=announce_float_stack ->
        let w=f_offset+(z-announce_float_stack)*8
        in  Dlist.replace ins (Opcode(F_Load(Double, frame_pointer, Imm13(w), r)))
    | Opcode(F_Store(Double, r, frame_pointer, Imm13(z))) when z>=announce_float_stack ->
        let w=f_offset+(z-announce_float_stack)*8
        in  Dlist.replace ins (Opcode(F_Store(Double, r, frame_pointer, Imm13(w))))
    | _ -> () ) sequence
(*}}}*);;

let generation arbre var_size identifier nb_args = 
(*{{{*)
  (*Printf.fprintf stderr "id = %s, var = %d, args = %d\n" identifier var_size nb_args;*)
  let procname = "proc__" ^ identifier
  and frame_label = "frame__" ^ identifier
  in
try

let top_of_frame_size = 8 in
let test,i_stack,f_stack,rd =
  if !Options.simplify then
  ( match arbre with
      Iexpr (tree) -> int_eval (Simplify.int_simpl tree) 0 
    | Fexpr (tree) -> float_eval (Simplify.float_simpl tree) 0
    | Vexpr (tree) -> let a,b,c = void_eval (Simplify.void_simpl tree) in a,b,c,0
    | Bexpr (tree) -> int_eval (Simplify.int_simpl (bool_to_int tree)) 0
    | Pexpr (_, tree, _) -> ptr_eval (Simplify.ptr_simpl tree) 0
  ) else
  ( match arbre with
      Iexpr (tree) -> int_eval (tree) 0 
    | Fexpr (tree) -> float_eval (tree) 0
    | Vexpr (tree) -> let a,b,c = void_eval (tree) in a,b,c,0
    | Bexpr (tree) -> int_eval (bool_to_int tree) 0
    | Pexpr (_, tree, _) -> ptr_eval (tree) 0
  )
in
let i_stack_size = max 0 ((i_stack - nb_hardware_registers)*4)
and f_stack_size = max 0 ((f_stack - nb_hardware_fregisters)*8)
in
let i_stack_offset = top_of_frame_size+var_size+i_stack_size
in
let f_stack_offset = ((i_stack_offset + f_stack_size + 4)/8)*8
in let total_frame_size = f_stack_offset + bottom_of_frame_size
in 
( if total_frame_size > 4096
  then Error.codegen_error (Printf.sprintf
    "Stack frame too big (%d > 4096) [top=%d, var=%d, istack=%d, fstack=%d, bot=%d]"
    total_frame_size top_of_frame_size var_size i_stack_size f_stack_size bottom_of_frame_size ) );

let 
toto = Dlist.singleton (Opcode(Breakpoint)) in
Dlist.insert_before
  (Opcode(Triadic(Save, false, stack_pointer, Imm13(-total_frame_size), stack_pointer))) toto;

for i=0 to (min (pred nb_args) 5)
do
  Dlist.insert_before (Opcode(Store(Word, (input_registers+i), frame_pointer, Imm13(parameters_offset+4*i)))) toto
done;

Dlist.insert_before (Opcode(Sethi((Hi_l frame_label), scratch_register))) toto;
Dlist.insert_before (Opcode(Load(Word, scratch_register, (Lo_l frame_label), overflow_ret_register))) toto;
Dlist.insert_before (Opcode(Store(Word, overflow_ret_register, frame_pointer, (Imm13 save_old_frame_offset)))) toto;
Dlist.insert_before (Opcode(Store(Word, frame_pointer, scratch_register, (Lo_l frame_label)))) toto;

Dlist.paste_before test toto;

Dlist.insert_before (Opcode(Sethi((Hi_l frame_label), scratch_register))) toto;
Dlist.insert_before (Opcode(Load(Word, frame_pointer, (Imm13 save_old_frame_offset), overflow_ret_register))) toto;
Dlist.insert_before (Opcode(Store(Word, overflow_ret_register, scratch_register, (Lo_l frame_label)))) toto;

(
match arbre with
  Fexpr(_) ->
    if (rd <> 0) then 
    ( Dlist.insert_before (Opcode(F_Diadic(F_Move, rd, 0))) toto;
      Dlist.insert_before (Opcode(F_Diadic(F_Move, (succ rd), 1))) toto; ()
    );
    Dlist.insert_before (Opcode(Ret)) toto;
    Dlist.insert_before (Opcode(Triadic(Restore, false, 0, Reg(0), return_fregister))) toto

| _ ->
    Dlist.insert_before (Opcode(Ret)) toto;
    Dlist.insert_before (Opcode(Triadic(Restore, false, rd, Reg(0), return_register))) toto
);


map_temp_stack (-i_stack_offset) (-f_stack_offset) toto;

(* Cette passe est indispensable pour renommer les labels dupliqus
   qui apparaissent quand des sauts gnrs sur une mme condition
   sont envoys sur le mme label. *)
Sparc.remove_duplicates toto;

Sparc.peephole toto;

print_string ("\t.local\t" ^ frame_label ^ "\n\t\t.common " ^ frame_label ^ ", 4, 4\n");

print_string (".section\t\".text\"\n\t.align 4\n\t.global\t" ^ procname
  ^ "\n\t.type\t" ^ procname ^ ", #function\n" ^ procname ^ " :\n" );
output_sequence stdout toto;
print_string (".Lend"^procname^" :\n\t.size " ^ procname ^ ", .Lend"^procname^" - " ^ procname ^ "\n")

with Failure(x) -> output_string stderr x; output_char stderr '\n'
(*}}}*);;

